import { Errors } from '../../lib/errors/__.js';
import { rootTypeNameToOperationName } from '../../lib/graphql.js';
import { isPlainObject } from '../../lib/prelude.js';
import { requestOrExecute } from '../0_functions/requestOrExecute.js';
import { Schema } from '../1_Schema/__.js';
import { readMaybeThunk } from '../1_Schema/core/helpers.js';
import { SelectionSet } from '../3_SelectionSet/__.js';
import * as CustomScalars from '../4_ResultSet/customScalars.js';
import { toDocumentString } from './document.js';
export const createPrefilled = (name, schemaIndex) => {
    // eslint-disable-next-line
    // @ts-ignore passes after generation
    return (input) => create({ ...input, name, schemaIndex });
};
export const create = (input_) => {
    // eslint-disable-next-line
    // @ts-ignore passes after generation
    const input = input_;
    /**
     * @remarks Without generation the type of returnMode can be `ReturnModeTypeBase` which leads
     * TS to think some errors below are invalid checks because of a non-present member.
     * However our implementation here needs to be generic and support all return modes
     * so we force cast it as such.
     */
    const returnMode = input.returnMode ?? `data`;
    const executeRootType = (context, rootTypeName) => async (selection) => {
        const rootIndex = context.schemaIndex.Root[rootTypeName];
        if (!rootIndex)
            throw new Error(`Root type not found: ${rootTypeName}`);
        // todo turn inputs into variables
        const documentString = SelectionSet.Print.rootTypeSelectionSet(context, rootIndex, 
        // @ts-expect-error fixme
        selection[rootTypeNameToOperationName[rootTypeName]]);
        // todo variables
        const result = await requestOrExecute({ schema: input.schema, document: documentString });
        // todo optimize
        // 1. Generate a map of possible custom scalar paths (tree structure)
        // 2. When traversing the result, skip keys that are not in the map
        // todo rename Result.decode
        const dataDecoded = CustomScalars.decode(rootIndex, result.data);
        return { ...result, data: dataDecoded };
    };
    const executeRootTypeField = (context, rootTypeName, key) => {
        return async (argsOrSelectionSet) => {
            const type = readMaybeThunk(
            // eslint-disable-next-line
            // @ts-ignore excess depth error
            Schema.Output.unwrapToNamed(readMaybeThunk(input.schemaIndex.Root[rootTypeName]?.fields[key]?.type)));
            if (!type)
                throw new Error(`${rootTypeName} field not found: ${String(key)}`); // eslint-disable-line
            // @ts-expect-error fixme
            const isSchemaScalarOrTypeName = type.kind === `Scalar` || type.kind === `typename`; // todo fix type here, its valid
            const isSchemaHasArgs = Boolean(context.schemaIndex.Root[rootTypeName]?.fields[key]?.args);
            const documentObject = {
                [rootTypeNameToOperationName[rootTypeName]]: {
                    [key]: isSchemaScalarOrTypeName
                        ? isSchemaHasArgs && argsOrSelectionSet ? { $: argsOrSelectionSet } : true
                        : argsOrSelectionSet,
                },
            };
            const result = await executeRootType(context, rootTypeName)(documentObject);
            const resultHandled = handleReturn(context.schemaIndex, result, returnMode);
            if (resultHandled instanceof Error)
                return resultHandled;
            return returnMode === `data` || returnMode === `dataAndErrors` || returnMode === `successData`
                // @ts-expect-error make this type safe?
                ? resultHandled[key]
                : resultHandled;
        };
    };
    const createRootTypeMethods = (context, rootTypeName) => new Proxy({}, {
        get: (_, key) => {
            if (typeof key === `symbol`)
                throw new Error(`Symbols not supported.`);
            // todo We need to document that in order for this to 100% work none of the user's root type fields can end with "OrThrow".
            const isOrThrow = key.endsWith(`OrThrow`);
            if (key.startsWith(`$batch`)) {
                return async (selectionSetOrIndicator) => {
                    const resultRaw = await executeRootType(context, rootTypeName)({
                        [rootTypeNameToOperationName[rootTypeName]]: selectionSetOrIndicator,
                    });
                    const result = handleReturn(context.schemaIndex, resultRaw, returnMode);
                    if (isOrThrow && result instanceof Error)
                        throw result;
                    // todo consolidate
                    // @ts-expect-error fixme
                    if (isOrThrow && returnMode === `graphql` && result.errors && result.errors.length > 0) {
                        throw new Errors.ContextualAggregateError(`One or more errors in the execution result.`, {}, 
                        // @ts-expect-error fixme
                        result.errors);
                    }
                    return result;
                };
            }
            else {
                const fieldName = isOrThrow ? key.slice(0, -7) : key;
                return async (argsOrSelectionSet) => {
                    const result = await executeRootTypeField(context, rootTypeName, fieldName)(argsOrSelectionSet); // eslint-disable-line
                    if (isOrThrow && result instanceof Error)
                        throw result;
                    // todo consolidate
                    // eslint-disable-next-line
                    if (isOrThrow && returnMode === `graphql` && result.errors.length > 0) {
                        throw new Errors.ContextualAggregateError(`One or more errors in the execution result.`, {}, 
                        // eslint-disable-next-line
                        result.errors);
                    }
                    return result;
                };
            }
        },
    });
    // @ts-expect-error ignoreme
    const client = {
        raw: async (input2) => {
            return await requestOrExecute({
                ...input2,
                schema: input.schema,
            });
        },
        rawOrThrow: async (input2) => {
            const result = await requestOrExecute({
                ...input2,
                schema: input.schema,
            });
            // todo consolidate
            if (result.errors && result.errors.length > 0) {
                throw new Errors.ContextualAggregateError(`One or more errors in the execution result.`, {}, result.errors);
            }
            return result;
        },
    };
    if (input.schemaIndex) {
        const schemaIndex = input.schemaIndex;
        const context = {
            schemaIndex,
            config: {
                returnMode,
            },
        };
        Object.assign(client, {
            document: (documentObject) => {
                const run = async (operationName) => {
                    // 1. if returnMode is successData OR using orThrow
                    // 2. for each root type key
                    // 3. filter to only result fields
                    // 4. inject __typename selection
                    // if (returnMode === 'successData') {
                    //   Object.values(documentObject).forEach((rootTypeSelection) => {
                    //     Object.entries(rootTypeSelection).forEach(([fieldExpression, fieldValue]) => {
                    //       if (fieldExpression === 'result') {
                    //         // @ts-expect-error fixme
                    //         fieldValue.__typename = true
                    //       }
                    //     })
                    //   })
                    // }
                    // todo this does not support custom scalars
                    const documentString = toDocumentString(context, documentObject);
                    const result = await requestOrExecute({
                        schema: input.schema,
                        document: documentString,
                        operationName,
                        // todo variables
                    });
                    return handleReturn(schemaIndex, result, returnMode);
                };
                return {
                    run,
                    runOrThrow: async (operationName) => {
                        const documentString = toDocumentString({
                            ...context,
                            config: {
                                ...context.config,
                                returnMode: `successData`,
                            },
                        }, documentObject);
                        const result = await requestOrExecute({
                            schema: input.schema,
                            document: documentString,
                            operationName,
                            // todo variables
                        });
                        // todo refactor...
                        const resultReturn = handleReturn(schemaIndex, result, `successData`);
                        return returnMode === `graphql` ? result : resultReturn;
                    },
                };
            },
            query: createRootTypeMethods(context, `Query`),
            mutation: createRootTypeMethods(context, `Mutation`),
            // todo
            // subscription: async () => {},
        });
    }
    return client;
};
const handleReturn = (schemaIndex, result, returnMode) => {
    switch (returnMode) {
        case `dataAndErrors`:
        case `successData`:
        case `data`: {
            if (result.errors && result.errors.length > 0) {
                const error = new Errors.ContextualAggregateError(`One or more errors in the execution result.`, {}, result.errors);
                if (returnMode === `data` || returnMode === `successData`)
                    throw error;
                return error;
            }
            if (returnMode === `successData`) {
                if (!isPlainObject(result.data))
                    throw new Error(`Expected data to be an object.`);
                const schemaErrors = Object.entries(result.data).map(([rootFieldName, rootFieldValue]) => {
                    // todo this check would be nice but it doesn't account for aliases right now. To achieve this we would
                    // need to have the selection set available to use and then do a costly analysis for all fields that were aliases.
                    // So costly that we would probably instead want to create an index of them on the initial encoding step and
                    // then make available down stream. Also, note, here, the hardcoding of Query, needs to be any root type.
                    // const isResultField = Boolean(schemaIndex.error.rootResultFields.Query[rootFieldName])
                    // if (!isResultField) return null
                    // if (!isPlainObject(rootFieldValue)) return new Error(`Expected result field to be an object.`)
                    if (!isPlainObject(rootFieldValue))
                        return null;
                    const __typename = rootFieldValue[`__typename`];
                    if (typeof __typename !== `string`)
                        throw new Error(`Expected __typename to be selected and a string.`);
                    const isErrorObject = Boolean(schemaIndex.error.objectsTypename[__typename]);
                    if (!isErrorObject)
                        return null;
                    // todo extract message
                    return new Error(`Failure on field ${rootFieldName}: ${__typename}`);
                }).filter((_) => _ !== null);
                if (schemaErrors.length === 1)
                    throw schemaErrors[0];
                if (schemaErrors.length > 0) {
                    const error = new Errors.ContextualAggregateError(`Two or more schema errors in the execution result.`, {}, schemaErrors);
                    throw error;
                }
            }
            return result.data;
        }
        default: {
            return result;
        }
    }
};
//# sourceMappingURL=client.js.map